
/*
 * Copyright (C) Igor Sysoev
 */


#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_http.h>
#include <ldap.h>

#define NGX_HTTP_AUTH_BUF_SIZE  2048

static ngx_conf_enum_t  ngx_http_auth_ldap_require[] = {
    { ngx_string("off"), 0 },
    { ngx_string("ldap-user"), 1 },
    { ngx_null_string, 0 }
};

#define NGX_HTTP_AUTH_LDAP_NONE 0
#define NGX_HTTP_AUTH_LDAP_VALID_USER 1

typedef struct {
    ngx_str_t    uri;
    ngx_uint_t   version;
} ngx_http_auth_ldap_main_conf_t;


typedef struct {
    ngx_str_t    realm;
    ngx_str_t    dn;
    ngx_str_t    attribute;
    ngx_uint_t   require;
    ngx_str_t   bind_dn;
    ngx_str_t   bind_passwd;
} ngx_http_auth_ldap_loc_conf_t;


static ngx_int_t ngx_http_auth_ldap_handler(ngx_http_request_t *r);
static ngx_int_t ngx_http_auth_ldap_set_realm(ngx_http_request_t *r, ngx_str_t *realm);
static ngx_int_t ngx_http_auth_ldap_init_module(ngx_cycle_t *cycle);
static void *ngx_http_auth_ldap_create_main_conf(ngx_conf_t *cf);
static void *ngx_http_auth_ldap_create_loc_conf(ngx_conf_t *cf);
static char *ngx_http_auth_ldap_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child);
static ngx_int_t ngx_http_auth_ldap_init(ngx_conf_t *cf);
static char *ngx_http_auth_ldap(ngx_conf_t *cf, void *post, void *data);
static char *ngx_http_auth_ldap_off(ngx_conf_t *cf, void *post, void *data);
static char *ngx_http_auth_ldap_uri(ngx_conf_t *cf, void *post, void *data);



static ngx_conf_post_handler_pt  ngx_http_auth_ldap_p = ngx_http_auth_ldap;
static ngx_conf_post_handler_pt  ngx_http_auth_ldap_off_p = ngx_http_auth_ldap_off;
static ngx_conf_post_handler_pt  ngx_http_auth_ldap_uri_p = ngx_http_auth_ldap_uri;


static ngx_command_t  ngx_http_auth_ldap_commands[] = {

    { ngx_string("auth_ldap_uri"),
      NGX_HTTP_MAIN_CONF|NGX_CONF_TAKE1,
      ngx_conf_set_str_slot,
      NGX_HTTP_MAIN_CONF_OFFSET,
      offsetof(ngx_http_auth_ldap_main_conf_t, uri),
      &ngx_http_auth_ldap_uri_p },
      
    { ngx_string("auth_ldap_version"),
      NGX_HTTP_MAIN_CONF|NGX_CONF_TAKE1,
      ngx_conf_set_num_slot,
      NGX_HTTP_MAIN_CONF_OFFSET,
      offsetof(ngx_http_auth_ldap_main_conf_t, version),
      NULL },

    { ngx_string("auth_ldap"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_HTTP_LMT_CONF
                        |NGX_CONF_TAKE1,
      ngx_conf_set_str_slot,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_auth_ldap_loc_conf_t, realm),
      &ngx_http_auth_ldap_p },

    { ngx_string("auth_ldap_bind_dn"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_HTTP_LMT_CONF
                        |NGX_CONF_TAKE1,
      ngx_conf_set_str_slot,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_auth_ldap_loc_conf_t, bind_dn),
      &ngx_http_auth_ldap_off_p },

    { ngx_string("auth_ldap_bind_passwd"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_HTTP_LMT_CONF
                        |NGX_CONF_TAKE1,
      ngx_conf_set_str_slot,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_auth_ldap_loc_conf_t, bind_passwd),
      NULL },

    { ngx_string("auth_ldap_dn"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_HTTP_LMT_CONF
                        |NGX_CONF_TAKE1,
      ngx_conf_set_str_slot,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_auth_ldap_loc_conf_t, dn),
      NULL },

    { ngx_string("auth_ldap_attribute"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_HTTP_LMT_CONF
                        |NGX_CONF_TAKE1,
      ngx_conf_set_str_slot,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_auth_ldap_loc_conf_t, attribute),
      NULL },

    { ngx_string("require"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_HTTP_LMT_CONF
                        |NGX_CONF_TAKE1,
      ngx_conf_set_enum_slot,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_auth_ldap_loc_conf_t, require),
      &ngx_http_auth_ldap_require },
      
    ngx_null_command
};


static ngx_http_module_t  ngx_http_auth_ldap_module_ctx = {
    NULL,                                  /* preconfiguration */
    ngx_http_auth_ldap_init,              /* postconfiguration */

    ngx_http_auth_ldap_create_main_conf,   /* create main configuration */
    NULL,                                  /* init main configuration */

    NULL,                                  /* create server configuration */
    NULL,                                  /* merge server configuration */

    ngx_http_auth_ldap_create_loc_conf,   /* create location configuration */
    ngx_http_auth_ldap_merge_loc_conf     /* merge location configuration */
};


ngx_module_t  ngx_http_auth_ldap_module = {
    NGX_MODULE_V1,
    &ngx_http_auth_ldap_module_ctx,       /* module context */
    ngx_http_auth_ldap_commands,          /* module directives */
    NGX_HTTP_MODULE,                       /* module type */
    NULL,                                  /* init master */
    ngx_http_auth_ldap_init_module,        /* init module */
    NULL,                                  /* init process */
    NULL,                                  /* init thread */
    NULL,                                  /* exit thread */
    NULL,                                  /* exit process */
    NULL,                                  /* exit master */
    NGX_MODULE_V1_PADDING
};

static int ngx_http_auth_ldap_bind(LDAP* ldap, char *dn, ngx_str_t* passwd)
{
    struct berval cred;
    cred.bv_len = passwd->len;
    cred.bv_val = (char*)(passwd->data);
    return ldap_sasl_bind_s( ldap, dn, LDAP_SASL_SIMPLE, &cred, NULL, NULL, NULL );
}

static ngx_int_t ngx_http_auth_ldap_handler(ngx_http_request_t *r)
{
    ngx_int_t                        rc;
    ngx_http_auth_ldap_loc_conf_t   *conf;
    ngx_http_auth_ldap_main_conf_t   *mconf;
    u_char                          *p;
    LDAP                            *ldap;
    LDAPMessage                     *msg;
    LDAPMessage                     *entry;
    char*                            dn;
    int                              deny;
    ngx_str_t                        filter;

    deny = 0;
    
    conf = ngx_http_get_module_loc_conf(r, ngx_http_auth_ldap_module);

    if ( conf->require == 0 ) {
        return NGX_OK;
    }
    
    rc = ngx_http_auth_basic_user(r);

    if (rc == NGX_DECLINED) {
        return ngx_http_auth_ldap_set_realm(r, &conf->realm);
    }

    if (rc == NGX_ERROR) {
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
    
    mconf = ngx_http_get_module_main_conf(r, ngx_http_auth_ldap_module);
    
    ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,"ngx_http_auth_ldap_handler: uri:%V", &(mconf->uri));
    
    rc = ldap_initialize(&ldap, (char*)(mconf->uri.data));
    
    if ( rc ) {
        ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,"ngx_http_auth_ldap_handler initialize: %s", ldap_err2string(rc));
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
    
    /* set protocol version */
    rc = ldap_set_option(ldap, LDAP_OPT_PROTOCOL_VERSION, &(mconf->version));

    if ( rc ) {
        ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,"ngx_http_auth_ldap_handler set_option: %s", ldap_err2string(rc));
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
    
    if ( conf->bind_dn.len > 0 ) { /* bind to bind_dn */
        rc = ngx_http_auth_ldap_bind( ldap, (char*)(conf->bind_dn.data), &(conf->bind_passwd) );
        if ( rc ) {
            ngx_log_error(NGX_LOG_ERR, r->connection->log, rc, "ngx_http_auth_ldap_handler error in bind: %V %s", &(conf->bind_dn), ldap_err2string(rc));
            return NGX_HTTP_INTERNAL_SERVER_ERROR;
        }
    }
    
    /* compose filer */
    filter.len = 3 + conf->attribute.len + r->headers_in.user.len;
    filter.data = ngx_pcalloc(r->pool, filter.len + 1);
    if ( filter.data == NULL ) {
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
    p = filter.data;
    *p++ = '(';
    p = ngx_cpymem(p, conf->attribute.data, conf->attribute.len);
    *p++ = '=';
    p = ngx_cpymem(p, r->headers_in.user.data, r->headers_in.user.len);
    *p++ = ')';
    *p++ = 0;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0, "ngx_http_auth_ldap_handler %V", &filter);
    
    /* search */
    rc = ldap_search_ext_s( ldap, (char*)(conf->dn.data), LDAP_SCOPE_SUBTREE, (char*)filter.data, NULL, 0, NULL, NULL, NULL, 1, &msg );
    if ( rc ) {
        ngx_log_error(NGX_LOG_ERR, r->connection->log, rc, "ngx_http_auth_ldap_handler failed ldap_search");
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
    
    rc = ldap_count_entries( ldap, msg );
    if ( rc == 0 ) { /* user not found */
        ngx_log_error(NGX_LOG_INFO, r->connection->log, rc, "ngx_http_auth_ldap_handler user not found: %V", &(r->headers_in.user));
        deny = 1;
    } else {
        entry = ldap_first_entry( ldap, msg );
        dn = ldap_get_dn(ldap, entry);
    
        rc = ngx_http_auth_ldap_bind( ldap, dn, &(r->headers_in.passwd) );
        if ( rc ) {
            ngx_log_error(NGX_LOG_ERR, r->connection->log, rc, "ngx_http_auth_ldap_handler bind: %s", ldap_err2string(rc));
            deny = 1;
        } else {
            ngx_log_error(NGX_LOG_INFO, r->connection->log, rc, "ngx_http_auth_ldap_handler logged in: %s", dn);
        }
    
        /* clean up */
        ldap_memfree(dn);
    }
    /* free LDAP message */
    rc = ldap_msgfree( msg );
    
    rc = ldap_unbind_ext_s( ldap, NULL, NULL );
    if ( rc ) {
        ngx_log_error(NGX_LOG_ERR, r->connection->log, rc, "ngx_http_auth_ldap_handler failed ldap_search");
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
    
    if ( deny ) {
        return ngx_http_auth_ldap_set_realm(r, &conf->realm);
    }
    
    return NGX_OK;
}

static ngx_int_t ngx_http_auth_ldap_set_realm(ngx_http_request_t *r, ngx_str_t *realm)
{
    r->headers_out.www_authenticate = ngx_list_push(&r->headers_out.headers);
    if (r->headers_out.www_authenticate == NULL) {
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }

    r->headers_out.www_authenticate->hash = 1;
    r->headers_out.www_authenticate->key.len = sizeof("WWW-Authenticate") - 1;
    r->headers_out.www_authenticate->key.data = (u_char *) "WWW-Authenticate";
    r->headers_out.www_authenticate->value = *realm;

    return NGX_HTTP_UNAUTHORIZED;
}

static void *ngx_http_auth_ldap_create_main_conf(ngx_conf_t *cf)
{
    ngx_http_auth_ldap_main_conf_t *conf;
    
    conf = ngx_pcalloc(cf->pool, sizeof(ngx_http_auth_ldap_main_conf_t));
    if (conf == NULL) {
        return NGX_CONF_ERROR;
    }
    conf->version = 3; /* default */
    
    return conf;
}

static void* ngx_http_auth_ldap_create_loc_conf(ngx_conf_t *cf)
{
    ngx_http_auth_ldap_loc_conf_t  *conf;

    conf = ngx_pcalloc(cf->pool, sizeof(ngx_http_auth_ldap_loc_conf_t));
    if (conf == NULL) {
        return NGX_CONF_ERROR;
    }
    conf->require = NGX_CONF_UNSET;

    return conf;
}


static char* ngx_http_auth_ldap_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
{
    ngx_http_auth_ldap_loc_conf_t  *prev = parent;
    ngx_http_auth_ldap_loc_conf_t  *conf = child;

    ngx_conf_merge_str_value(conf->realm, prev->realm, "");
    ngx_conf_merge_str_value(conf->dn, prev->dn, "");
    ngx_conf_merge_str_value(conf->attribute, prev->attribute, "");
    ngx_conf_merge_uint_value(conf->require, prev->require, 0);
    ngx_conf_merge_str_value(conf->bind_dn, prev->bind_dn, "");
    ngx_conf_merge_str_value(conf->bind_passwd, prev->bind_passwd, "");

    return NGX_CONF_OK;
}

static ngx_int_t ngx_http_auth_ldap_init(ngx_conf_t *cf)
{
    ngx_http_handler_pt        *h;
    ngx_http_core_main_conf_t  *cmcf;

    cmcf = ngx_http_conf_get_module_main_conf(cf, ngx_http_core_module);

    h = ngx_array_push(&cmcf->phases[NGX_HTTP_ACCESS_PHASE].handlers);
    if (h == NULL) {
        return NGX_ERROR;
    }

    *h = ngx_http_auth_ldap_handler;

    return NGX_OK;
}

static char *ngx_http_auth_ldap_off(ngx_conf_t *cf, void *post, void *data)
{
    ngx_str_t   *value = data;
    
    if (ngx_strcmp(value->data, "off") == 0) {
        value->len = 0;
        value->data = (u_char *) "";
    }
    
    return NGX_CONF_OK;
    
}

static char *ngx_http_auth_ldap(ngx_conf_t *cf, void *post, void *data)
{
    ngx_str_t  *realm = data;

    size_t   len;
    u_char  *basic, *p;

    if (ngx_strcmp(realm->data, "off") == 0) {
        realm->len = 0;
        realm->data = (u_char *) "";

        return NGX_CONF_OK;
    }

    len = sizeof("Basic realm=\"") - 1 + realm->len + 1;

    basic = ngx_pnalloc(cf->pool, len);
    if (basic == NULL) {
        return NGX_CONF_ERROR;
    }

    p = ngx_cpymem(basic, "Basic realm=\"", sizeof("Basic realm=\"") - 1);
    p = ngx_cpymem(p, realm->data, realm->len);
    *p = '"';

    realm->len = len;
    realm->data = basic;

    return NGX_CONF_OK;
}

static char* ngx_http_auth_ldap_uri(ngx_conf_t *cf, void *post, void *data)
{
    ngx_str_t *uri = data;

    if(!ldap_is_ldap_url((char*)uri->data)) {
	ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "This is not a LDAP URI:\"%V\"\n",uri);
        return NGX_CONF_ERROR;
    }

    return NGX_CONF_OK;
}

static ngx_int_t ngx_http_auth_ldap_init_module(ngx_cycle_t *cycle)
{
    ngx_int_t                        rc;
    ngx_http_auth_ldap_main_conf_t   *mconf;
    LDAP                            *ldap;
    
    mconf = ngx_http_cycle_get_module_main_conf( cycle, ngx_http_auth_ldap_module );
    
    /* initialize ldap */
    rc = ldap_initialize(&ldap, (char*)(mconf->uri.data));
    
    if ( rc ) {
        ngx_log_error(NGX_LOG_ERR, cycle->log, 0,"ngx_http_auth_ldap_handler initialize: %s", ldap_err2string(rc));
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
    
    /* set protocol version */
    rc = ldap_set_option(ldap, LDAP_OPT_PROTOCOL_VERSION, &(mconf->version));

    if ( rc ) {
        ngx_log_error(NGX_LOG_ERR, cycle->log, 0,"ngx_http_auth_ldap_handler set_option: %s", ldap_err2string(rc));
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
    
    /* unbind */
    rc = ldap_unbind_ext_s(ldap, NULL, NULL);

    if ( rc ) {
        ngx_log_error(NGX_LOG_ERR, cycle->log, 0,"ngx_http_auth_ldap_handler unbind: %s", ldap_err2string(rc));
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
    
    return NGX_OK;
}
